/*
 * This is the first project assignment for CSE 190.
 * Created Jan 17, 96, bsy.
 *
 * Imagine that this is an electronic homework submission system,
 * where the server accepts your answers via the network, evaluates
 * your answers, and enters out your grade into a database.
 *
 * Your task is to get this network server to print out your (desired)
 * grade without solving the actual, very difficult problem set.
 *
 * Because UCSD system administrators requested that we do not actually
 * have students break into accounts, this is a simulation ONLY.
 * The simulated network server is contained in this file.  It requires
 * two additional functions to be provided by you -- the client_op
 * function and the client_answer function.
 *
 * The client_op function simulates network input from the client to
 * the server.  It returns 3 values, 2 of which is by modifying memory
 * pointed to by 2 input parameters.  The functional return value is
 * an OP code, specifying what operation is to be done by the server.
 * The 2 argument result pointers are to a length value and a buffer
 * pointer.  When the OP code is for an operation that takes additional
 * data as parameter, the server will copy the specified number of bytes from
 * the client-supplied buffer into its own buffer (simulating a network
 * read operation) before processing the data.
 *
 * The client_answer function simulates network output from the server
 * to the client.  It takes 3 input arguments, an OP code, a buffer
 * length, and a buffer pointer similar to before.  The client code
 * should copy any data out of the buffer as specified by the length.
 * The client code may not read pass the end of the buffer nor overwrite
 * the data that's there, since this is simulating a network I/O
 * operation and the client can not directly affect the server's memory
 * in this fashion.
 *
 *
 * Grading:
 *
 *  10 points -- provide client code which gets the server to
 *	enter an 'A' grade into the grade database for yourself
 *	with the server unmodified.
 *  30 points -- provide client code which gets the server to
 *	enter an 'A' grade as before after removing the first
 *	"homework problem" by changing the preprocessor directive
 *		#define	EASY	1
 *	below to
 *		#define	EASY	0
 *  20 points -- written:  identify other security issues in this code.
 *	(E.g., denial of service attacks.)
 *  40 points -- provide client code which causes the server to
 *	create a new file in the local directory named "Gotcha!"
 *
 * All code should be accompanied by documentation describing the security
 * flaw(s) exploited and a recommendation on how the flaw(s) should be fixed.
 *
 *
 * Ground rules:
 *
 * All source code in this file should remain untouched (except for EASY
 * preprocessor symbol mentioned above).
 * You must write code which uses the server's published interface -- see the
 * header file ass1.h -- to print out the grade that you "deserve".
 * Because this is a simulation, your "client" code shares address space
 * with the "server"; directly modifying memory via wild pointers is not
 * acceptable:  you may influence the "server" state only by calling
 * the functions in the published interface (but you are free to supply
 * arbitrary parameters).  Your "client" code can not print out the grade
 * by itself, of course, since it is supposed to be running in a separate
 * address space on a separate machine and can't access the grades database.
 *
 */

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <setjmp.h>
#include <stdlib.h>

/* MD5's defns */
#include <global.h>
#include <md5.h>

#include "assn3.h"

#define	GRADE_DB	"grade.db"

#define	EASY	1

/* server's public variables */

int	debug = 0;

/* server's private variables */

struct prob_tbl {
	void		*(*init_fn)(void);
	/* allocates and returns problem context */
	void		(*process_fn)(void *,unsigned short, const char *);
	/* processes a buffer full of data */
	unsigned int	(*eval_fn)(void *);
	/* frees context score for problem */
};

/* per connection data */
struct context {
	char		student_name[256];
	struct prob_tbl	*cur_prob;
	void		*prob_ctx;

	unsigned short	buffer_byte_count;
	char		buffer[1 << 14];
	unsigned int	score;			/* score so far */
};

static char		*prog_name;
static jmp_buf		next_student;

static struct {
	int	min_score; char *grade;
} grade_table[] = {
	0,	"F",
	65,	"D",
	70,	"C",
	75,	"C+",
	80,	"B",
	85,	"B+",
	90,	"A",
	95,	"A+",
};
#define	MAX_GRADE_LEN	3	/* incl NULL */

#if	EASY

struct stupid_problem_state {
	int	got_answer;
};

static void	*s_init(void)
{
	struct stupid_problem_state	*s;

	s = malloc(sizeof *s);
	if (s) s->got_answer = 0;
	return s;
}

static void	s_data(void		*state,
		       unsigned short	byte_count,
		       const char	*buf)
{
	struct stupid_problem_state	*s = state;
	unsigned long			ans;

	if (byte_count == sizeof ans) {
		memcpy(&ans,buf,sizeof ans);
		ans = ntohl(ans);
		if (ans == 362880) s->got_answer = 1;	/* 10! */
	}
}

static unsigned int	s_done(void		*state)
{
	struct stupid_problem_state	*s = state;
	int				score;

	score = s->got_answer ? 10 : 0;
	free(s);
	return score;
}

#endif
/* EASY */

static void	*h_init(void)
{
	MD5_CTX	*ctx;

	ctx = malloc(sizeof *ctx);
	if (!ctx) return 0;
	MD5Init(ctx);
	return (void *) ctx;
}

static void	h_data(void		*state,
		       unsigned short	byte_count,
		       const char	*buf)
{
	MD5_CTX	*ctx = state;
	MD5Update(ctx,(unsigned char *) buf,byte_count);
}

static void	print_checksum(const unsigned char	*cksum,
			       FILE			*iop)
{
	int	i;

	for (i = 0; i < 0x10; i++) {
		if (!(i & 0x7)) putc(' ',iop);
		fprintf(iop," %02x",cksum[i]);
	}
	putc('\n',iop);
}

/* a rather binary evaluation; no partial credits */
static unsigned int	h_done(void		*state)
{
	unsigned char		cksum[16];
	static unsigned char	desired[16] = {
		0x62, 0x73, 0x79, 0x20,
		0x77, 0x61, 0x73, 0x20,
		0x68, 0x65, 0x72, 0x65,
		0x2e, 0x2e, 0x2e, 0x21,
	};
	MD5_CTX			*ctx = state;
	int			diff;

	MD5Final(cksum,ctx);
	fprintf(stderr,"Client supplied data resulted in checksum:\n");
	print_checksum(cksum,stderr);
	fprintf(stderr,"Desired checksum:\n");
	print_checksum(desired,stderr);
	free(ctx);
	diff = memcmp(cksum,desired,sizeof cksum);
	return diff ? 0 :
#if	EASY
		80;
#else
		90;
#endif
	/* score for problem-set should adds up to 100 */
}

struct prob_tbl prob_tbl[] = {
	0,	0,	0,	/* 0: no problem 0 */
#if	EASY
	s_init,	s_data,	s_done,	/* 1: find value of 10 factorial */
#endif
	h_init,	h_data,	h_done,	/* 2: find MD5 preimage */
};

static void	enter_grade(const char	*name,
			    char	*grade)
{
	static FILE	*grade_db = 0;
	if (!grade_db) grade_db = fopen(GRADE_DB,"a");
	if (!grade_db) fprintf(stderr,"%s: cannot open grade database \"%s\"\n",prog_name,GRADE_DB);
	else fprintf(grade_db,"Student %s receives an %s for project 1\n",name,grade);
}

/* byte count should include the null termination */
static void	student_name(struct context	*ctx,
			     unsigned short	byte_count,
			     const char		*name)
{
	if (byte_count >= sizeof ctx->student_name) {
		fprintf(stderr,"Student name too long\n");
		enter_grade(name,"F");
		longjmp(next_student,1);
	}
	strncpy(ctx->student_name,name,byte_count);
	ctx->student_name[byte_count-1] = 0;
	/* make sure it is null terminated */
	if (strlen(ctx->student_name) != byte_count-1) {
		fprintf(stderr,"Mal-formed student name\n");
		enter_grade(name,"F");
		longjmp(next_student,1);
	}
	ctx->cur_prob = &prob_tbl[1];
	ctx->score = 10;
	/* get 10 points for getting name right */
}

static void	problem_number(struct context	*ctx,
			       unsigned short	byte_count,
			       const char	*buffer)
{
	unsigned short	num;

	if (byte_count != sizeof num) {
		fprintf(stderr,"Incorrect sized problem number\n");
		enter_grade(ctx->student_name,"F");
		longjmp(next_student,1);
	}
	memcpy(&num,buffer,sizeof num);
	num = ntohs(num);
	if (!num || num >= sizeof prob_tbl/sizeof prob_tbl[0]) {
		fprintf(stderr,"No such problem (%d)\n",num);
		enter_grade(ctx->student_name,"F");
		longjmp(next_student,1);
	}
	ctx->cur_prob = &prob_tbl[num];
	ctx->prob_ctx = (*ctx->cur_prob->init_fn)();
	if (!ctx->prob_ctx) {
		fprintf(stderr,"Problem context initialization error\n");
		/* abort connection, no score given */
		exit(1);
	}
}

static void	process_data(struct context	*ctx,
			     unsigned short	byte_count,
			     const char		*buffer)
{
	(*ctx->cur_prob->process_fn)(ctx->prob_ctx,byte_count,buffer);
}

static void	eval_answer(struct context	*ctx,
			    unsigned short	byte_count,
			    const char		*buffer)
{
	unsigned long	score;

	score = (*ctx->cur_prob->eval_fn)(ctx->prob_ctx);
	if (debug) {
		printf("eval_answer scores %lu\n",score);
	}
	ctx->score += score;
	if (debug) {
		printf("score so far is %lu\n",score);
	}
	ctx->prob_ctx = 0;
	score = htonl(score);
	client_answer(RESP_SCORE,sizeof(score),(const char *) &score);
}

struct op_tbl {
	int	multi_frag_ok;
	void	(*fn)(struct context *,unsigned short,const char *);
} op_tbl[] = {
	0,	student_name,		/* OP_STUDENT_ID */
	0,	problem_number,		/* OP_SET_PROBLEM_NUM */
	1,	process_data,		/* OP_MORE_DATA */
	0,	eval_answer,		/* OP_EVALUATE */
};

static void main_loop(void)
{
	int		op, prev_op = OP_DONE;
	char		*cli_buf, *grade;
	unsigned short	cli_len, frag, new_byte_count;
	struct context	ctx;
	int		i;
	unsigned long	total;
	char		grade_buf[sizeof(unsigned long) + MAX_GRADE_LEN];


	/*
	 * We "read" data from the "network" into an I/O buffer
	 * for processing.  Answers may arrive fragmented in
	 * multiple data "packets", and we "reassemble" them
	 * in our I/O buffer.  Furthermore, some questions may
	 * require answers that is longer than the maximum
	 * transmissible unit in length -- in that case, we
	 * assume that the problem handler has enough state
	 * saved in the problem context (see prob_tbl's init_fn)
	 * so as to be able to handle multiple buffers iteratively.
	 */
	if (setjmp(next_student)) return;
	ctx.buffer_byte_count = 0;
	while ((op = client_op(&cli_len,&cli_buf)) < OP_DONE) {
		if (debug) {
			printf("Got op %d, len %d, buf 0x%08lx\n",op,cli_len,(long) cli_buf);
		}
		if (op != prev_op && ctx.buffer_byte_count) {
			if (debug)
				printf("Flushing old data\n");
			(*op_tbl[prev_op].fn)(&ctx,ctx.buffer_byte_count,ctx.buffer);
			ctx.buffer_byte_count = 0;
		}
		prev_op = op;
		if (!op_tbl[op].multi_frag_ok) {
			if (debug) printf("not multi fragments\n");
			memcpy(ctx.buffer,cli_buf,cli_len);
			(*op_tbl[op].fn)(&ctx,cli_len,ctx.buffer);
		} else {
			new_byte_count = ctx.buffer_byte_count + cli_len;
			if (new_byte_count <= sizeof ctx.buffer) {
				if (debug) printf("fragment saved\n");
				memcpy(ctx.buffer + ctx.buffer_byte_count,cli_buf,cli_len);
				ctx.buffer_byte_count = new_byte_count;
			} else {
				frag = (sizeof ctx.buffer) - ctx.buffer_byte_count;
				if (debug) printf("frag cause buffer to fill, processing %d 0x%x bytes\n",frag,frag);
				memcpy(ctx.buffer + ctx.buffer_byte_count,cli_buf,frag);
				(*op_tbl[op].fn)(&ctx,sizeof ctx.buffer,ctx.buffer);
				memcpy(ctx.buffer,cli_buf + frag,cli_len - frag);
				ctx.buffer_byte_count = cli_len - frag;
			}
		}
	}
	/* calculate project/assignment grade */
	for (i = 0; i < sizeof grade_table/sizeof grade_table[0]; i++) {
		if (ctx.score >= grade_table[i].min_score)
			grade = grade_table[i].grade;
	}
	total = htonl(ctx.score);
	memcpy(grade_buf,&total,sizeof total);
	strcpy(grade_buf + sizeof total,grade);
	client_answer(RESP_FINAL_SCORE,sizeof grade_buf,grade_buf);
	enter_grade(ctx.student_name,grade);
}

int main(int	ac,
	 char	**av)
{
	int		opt;

	prog_name = strrchr(av[0],'/');
	if (!prog_name) prog_name = av[0]; else ++prog_name;
	while ((opt = getopt(ac,av,"d")) != EOF) switch (opt) {
	case 'd':	++debug;		break;
	default:
		fprintf(stderr,"%s:  illegal option %c\n",prog_name,opt);
		exit(1);
	}
	main_loop();
	printf("Simulation ends (no next student)\n");
	return 0;
}
